# -*- coding: utf-8 -*- #

""" Handle the pipeline manager tab. Allow to create and run pipelines.

Contains:
    Module:
        - iteration_table
        - node_controller
        - pipeline_editor
        - pipeline_manager_tab
        - process_library
        - process_mia

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################
