# -*- coding: utf-8 -*- # Character encoding, recommended
""" Make populse_mia accessible as a module.

For example by using python3 -m populse_mia.

"""
##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

# Populse_MIA imports
from populse_mia.main import main

if __name__ == '__main__':
    main()
