# -*- coding: utf-8 -*- #

""" Handle the data browser tab. Visualize, modify and search data.

Contains:
    Module:
      - advanced_search
      - count_table
      - data_browser
      - mini_viewer
      - modify_table
      - rapid_search

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################
