# -*- coding: utf-8 -*- #

""" User interface module, encompass the main window and its tabs
definition, as well as pop ups.

Contains:
    Module:
        - main_window
        - pop_ups
        - data_browser
        - pop_ups

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################
